/* This file contains the GOLD Maxflow Algorithm developed by 
   Cheriyan & Hagerup. For details look: 
   CH2806-8/89/0000/0118/$01.00(c)1989IEEE */

/* Copyright:
   This program was written by 
   
   Tamas Badics, 1991,
   Rutgers University, RUTCOR
   P.O.Box 5062
   New Brunswick, NJ, 08903
   e-mail: badics@rutcor.rutgers.edu
   
   The code may be used and modified for not-for-profit use.
   This notice must be remained.	
   ====================================================================*/

#include "_maxflow.h"
#include "macros.h"
#include "netio.h"

#ifdef DEBUG
#include "debug.h"
#endif

static int quiet;
int It, Big;

#ifdef COUNT
int Bigrel,Relab, Sat, Nonsat;
#endif
/*==================================================================*/
int maxflow(network *n, controll *cont)
                
                    
	 /* Finds the max-flow in the initialized 
		network n using the GOLD alg.  */
{
	vertex *v;
	int pf, sf, rf, last;
	int big = 1, stall = 0, was_change = 0;
	DOUBLE old_fl;   
	int br = 0;
	
	It = Big = 0;
	
#ifdef COUNT
	Bigrel = Relab = Sat = Nonsat = 0;
#endif
	last = 0;
	
	pf = cont->print_freq;
	sf = cont->stall_freq;
	rf = cont->relab_freq;
	
	quiet = cont->quiet;
	Gap = cont->gap;
	
	while (!IS_EMPTY_QUEUE(Qv)){
		It++;
		
		v = (vertex *)dequeue(Qv);
		
		if (v == NULL)
		  print_err("Null dequeued", 1);
		
		old_fl = Sink->excess;
		
		discharge(v);
		
		if (v->excess > 0 && v->label != Inf)
		  enqueue(Qv, v);
		
		/* -------------------------------------
		   Big_relabel heuristics 
		   -------------------------------------
		   */
		
		if (Sink->excess == old_fl)
		  stall++;
		else{
			old_fl = Sink->excess;
			was_change = 1;
			stall = 0;
		}
		
		if ((was_change && sf && stall >= sf)
			|| (rf && rf <= It - last)
			|| (Gap && Big)){ 
			
			if (pf != 0){
				print_report();
				PRINT( quiet, ("BIG_RELABEL..%d\n", ++br));
			}
			
			stall = 0;
			was_change = 0;
			last = It;
			Big = 0;
			
			if (big_relabel(n, big)){
				if (pf != 0) PRINT( quiet, ("MINCUT!\n"));
				big = 0;
			}
		}
		
		if (pf > 0 && It%pf == 0)
		  print_report();
	}
	
	PRINT(quiet, ("========== END ============\n\n"));
	
#ifdef COUNT
	PRINT(quiet, ("Bigrel: %d Select: %d Relab: %d Sat: %d Nonsat: %d\n"
				  , Bigrel, It, Relab, Sat, Nonsat));
#endif
	
	n->maxflow = Sink->excess;
	return 1;
}

/*==================================================================*/
void pushrelabel(vertex *v)
{
	int vl;
	edge *e;
	vertex * w;
	
	e = *v->current;
	w = OTHER(v, e);
	
	if (v->label == w->label + 1 && RESCAP(v, e) > 0){
		if (w->excess == 0 && w != Source && w != Sink)
		  enqueue(Qv, w);
		push(v, e);
	}
	else if (v->current == LAST_EDGE(v)){
		v->current = v->edgelist;
		
		if (Gap && v->label < Vertnum){ 
			(Labarr[(vl = v->label)])--;
			relabel(v);
			
			if (v->label < Vertnum){ 
				(Labarr[v->label])++;
				if (v->label > Labarr[0])
				  Labarr[0] = v->label;
			}else{
				if (vl == Labarr[0] && Labarr[vl] == 0)
				  (Labarr[0])--;
			}
			
			if (Labarr[vl] == 0 && vl < Labarr[0])
			  Big = 1;
			else
			  Big = 0;
			
		}else
		  relabel(v);
	}else
	  v->current++;
}
/*==================================================================*/
void discharge(vertex *v)
{
	int d;
	
	d = v->label;
	while (v->label == d && v->excess > 0)
	  pushrelabel(v);
}
/*==================================================================*/
void push(vertex *v, edge *e)
{
	DOUBLE c;
	vertex * w;
	
	c = RESCAP(v, e);
	c = MIN2(c, v->excess);
	
#ifdef COUNT
	if (c == v->excess)
	  Nonsat++;
	else
	  Sat++;
#endif	
	
	v->excess -= c;
	
	if (OUT_EDGE(v, e)){
		e->flow += c;
		(w = e->to)->excess += c;
	}else{
		e->flow -= c;
		(w = e->from)->excess += c;
	}
}
/*==================================================================*/
void relabel(vertex *v)
{
	int i, l;
	edge * e, ** ep;
	vertex * w;
	
#ifdef COUNT
	Relab++;
#endif	
	
	for (v->label = Inf, i = 0, ep = v->edgelist
		 ; i < v->degree; i++){
		
		e = ep[i];
		w = OTHER(v, e);
		
		if ((l = w->label) < v->label && RESCAP(v, e) > 0)
		  v->label = l;
	}
	v->label += 1;
}

/*=================================================================*/
void print_report(void)
{
#ifdef COUNT	
	PRINT( quiet, ("Select: %9d\nRelab:  %9d\nSat:    %9d\nNonsat: %9d\n------------- ", It, Relab, Sat, Nonsat));
#endif
	PRINT( quiet, ("Flow: %d\n", (int)(Sink->excess)));
}
