/*  edge.h - Definition of the edge class for the DISTRIBUTED max flow
 *           Goldberg-Tarjan algorithm.
 *
 *  programmer:     Austin Armbruster
 *  creation date:  13 October 2000
 *  last modified:
 */

#ifndef _EDGE_DIST_H_
#define _EDGE_DIST_H_

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "mflib.h"

class edge
{
 public:

  // CONSTRUCTORS and DESTRUCTOR for edge
  edge(int direction, int id, int distance, char* serverName, int flowPort, 
       int inUpperLimit, int inLowerLimit, int inUpperCap, int inLowerCap=0, 
       int inCost=1);
  /* Creates a new edge with the given information
   * PRE:  direction is either COMINGFROM or GOINGTO.  The 
   *          direction is looking from the vertex that the edge 
   *          is being listed with.  For example if A-->B, the 
   *          edge would be listed using GOINGTO on the A vertex 
   *          and COMINGFROM on the B vertex.
   *       id is the id of the other vertex, i.e. not the one 
   *          this edge is listed with.
   *       serverName is the host name of the server where the 
   *          other vertex is located.
   *       flowPort is the port number on which the other vertex 
   *          is listening for messages to accept flow.
   *       distancePort is the port number on which the other 
   *          vertex is listening for messages that adjust 
   *          distances.
   *       inUpperCap is the upper capacity of the edge.
   *       inLowerCap is the lower capacity of the edge.  If not 
   *          provided, it will be 0.
   *       inCost is the cost, or weighting, of the edge.  If not
   *          provided, it will be 1.
   * POST: The edge will be instantiated using the given 
   *          information.
   ***************************************************************/

  // MODIFICATION MEMBER FUNCTIONS for edge

  void modifyFlow(int inDelta, int inDirection, bool mfDirection);
  /* modifyFlow changes the amount of flow across the edge by the 
   *     amount of inDelta.
   * PRE:  if the flow is being pushed along the direction of the 
   *          edge, then m_flow+inDelta <= m_upperCapacity.
   *       if the flow is being pushed in the opposite direction 
   *          of the edge, then m_flow-inDelta >= m_lowerCapacity.
   * POST: The value of m_flow will be changed appropriatly.
   ***************************************************************/   

  void setovDistance(int arg);
  /* setovDistance sets the distance for the other vertex 
   *    attached to this edge.
   * PRE:  NONE
   * POST: m_ovDistance = arg
   ***************************************************************/   
  void setEdge();

  // CONSTANT MEMBER FUNCTION for edge

  int upperCapacity() const;
  /* upperCapacity returns the value of m_upperCapacity
   * PRE:  NONE
   * POST: m_upperCapacity is returned
   ***************************************************************/

  int rf(bool mfDirection) const;
  /* rf returns true if there is still residual capacity.
   * PRE:  NONE
   * POST: rf returns rhe amount of residual capacity in the edge.
   ***************************************************************/

  int ovDistance() const;
  /* ovDistance returns the local value of the distance of the 
   *     other vertex connected to by this edge.
   * PRE:  NONE
   * POST: The value stored in m_ovDistance is returned.
   ***************************************************************/

  int getMaxLimit() const;

  sockaddr_in getAddr() const;
  /* Returns the socket address to the port of the vertex
   *   on the other end of the edge is returned.
   * PRE:  NONE
   * POST: the socket address to the port of the vertex 
   *         on the other end of the edge is returned.
   ***************************************************************/

  int ovId() const;
  /* ovId returns the id number of the other vertex attached to
   *     this edge for the.
   * PRE:  NONE
   * POST: returns m_ovId
   ***************************************************************/

  int getFlow() const;
  /* getFlwo returns the amount of flow on this edge
   * PRE:  NONE
   * POST: returns m_flow
   ***************************************************************/

  int direction() const;
  /* returns the direction of the edge.  Recall that this is 
   *    oriented with respect to the vertex that lists the edge.
   * PRE:  NONE
   * POST: returns m_direction
   ***************************************************************/

  static const int GOINGTO = 1;
  static const int COMINGFROM = -1;
  
  friend bool operator<( edge& s1, edge& s2);

 private:
  int m_flow;
  int m_upperLimit, m_lowerLimit;
  int m_upperCapacity;
  int m_lowerCapacity;
  int m_cost;
  bool m_set;

  int m_direction;

  int m_ovDistance;
  int m_ovId;
  sockaddr_in m_ovAddr;
};

#endif
