/*  vertex.h - Definition file for the vertex class used in the
 *             DISTRIBUTED max flow algorithm using Goldberg and Tarjan's
 *             asynchronous method.
 *
 *  programmer:
 *  creation date:  14 October 2000
 *  last modified:  
 */

#ifndef _VERTEX_H_
#define _VERTEX_H_

#include <list>
#include <map>
#include <vector>
#include <string>

#include "mflib.h"
#include "message.h"
#include "edge.h"
#include "acceptFlowData.h"
#include "sendFlowData.h"

class vertex
{
 public:

  // CONSTRUCTORS and DESTRUCTOR for vertex
  vertex(int id, char* fileName, bool verbose = false);
  /* The default constructor of the vertex
   * PRE:  id must be the number of this vertex in the input file.
   *       fileName must be the name, full path if necessary, of 
   *          the input file.
   *       verbose should be true to get all output comments
   * POST: a vertex is created that is ready to accept new edges
   *****************************************************************/

  ~vertex();
  /* The destructor will release all memory associated with the 
   *     vertex
   * PRE:  NONE
   * POST: All memory associated with vertex will be released back 
   *       to the heap.
   *****************************************************************/

  // MODIFICATION MEMBER FUNCTIONS for vertex


  void pulse();
  /* Pulse attempts to push flow to other acceptable nodes.
   * PRE:  NONE
   * POST: Flow is pushed to other nodes that were willing to 
   *       accept the flow.  If a vertex rejected the flow, a new
   *       distance was recorded for that vertex.  If a new label 
   *       was needed for this vertex, the new label was recorded, 
   *       and that label was broadcast to the neighboring 
   *       vertices.
   *****************************************************************/

  void push(edge& ce);
  /* Push determines how much flow should go to the given edge, and
   *     then tries to send that to the vertex.
   * PRE:  This vertex is active.
   *       ce's residual flow is positive.
   *       This vertex's distance is equal to the distance of the 
   *          vertex on the other end of ce plus 1, at least to 
   *          current local settings.
   * POST: This vertex's excess is reduced by the amount that is 
   *       trying to be pushed to the other vertex.  Another thread 
   *       is created to send the flow to the other vertex and wait 
   *       for a response.  The thread id of the new thread is 
   *       added to m_tidList.
   *****************************************************************/

  void receivedResponse(sendFlowData& sfd);

  void relabel();
  /* relabel determines what the new distance label for the vertex 
   *     should be.
   * PRE:  NONE
   * POST: The vertex label was changed to the new proper labeling.
   *****************************************************************/

  // CONSTANT MEMBER FUNCTIONS for vertex

  inline bool active();
  /* Active determines if the vertex is active or not.
   * PRE:  The mutex m_excessLock is locked.
   * POST: Active returns true if m_excess > 0; otherwise, it 
   *       returns false.
   *****************************************************************/

  bool properDistance(edge& ce);
  /* properDistance determines if the vertex is the proper 
   *     distance fromthe other vertex on ce to push flow.
   * PRE:  NONE
   * POST: Returns true is m_distance is 1 plus the distance of the 
   *          other vertex, according to local data.  Otherwise, 
   *          false is returned.
   *****************************************************************/

  void listenSocketSetup(int& port);

  void acceptFlow(message& msg, int& ns);
  /* accpetFlow is run in a separate thread to continuously monitor 
   *     for new and handle messages from other vertices trying to 
   *     push flow to this vertex.
   * PRE:  NONE
   * POST: Currently, the thread does not end.
   *****************************************************************/

  void acceptDistance(message& msg, int& ns);
  /* acceptNewDistances runs in a separate thread and listens for 
   *     the broadcasts of distances from other vertices.
   * PRE:  NONE
   * POST: The broadcasts of distances from other vertices were 
   *          read, and the distance was stored locally.
   *****************************************************************/

  void sendDistance(edge& ce);
  /* sendDistance sends the distance of this vertex to the other
   *     vertex attached to the edge ce.
   * PRE:  ce must point to an edge.
   *       ce must be attached to this vertex.
   * POST: The distance of this vertex is sent to the other vertex 
   *          attached to the edge ce.
   *****************************************************************/

  void maxFlow();
  /* maxFlow tries to pulse out excess from this vertex while it is 
   *    active.  It waits for more flow if it is not active and 
   *    the algorithm has not terminated.  It exits once the 
   *    algorithm has terminated.
   * PRE:  NONE
   * POST: All received flow has been pulsed out.
   *       The entire algorithm has terminated.
   *****************************************************************/

  void controllerListenerSetup(const char* server, int port);
  void controllerListener(message& msg, int ns);
  /* terminationListener waits for this vertex to become ready to 
   *    receive flow and distances.  It then sends a message to the 
   *    controller that it is ready.  It waits for a message that
   *    states the other vertices are also ready.  It notes that
   *    the network is ready and notifies anyone waiting on that 
   *    condition.  It then listens for a message that the algorithm
   *    has terminated.
   * PRE:  NONE
   * POST: The algorithm has terminated.
   *****************************************************************/

  void vertex::acceptTerm(message& msg);


  void sendTerm(acceptFlowData& afd);

  /* sendTermWeight takes care of setting up the socket and sending
   *    the message that tells the controller that this vertex 
   *    has completed pulsing all of it's excess.  
   * PRE:  arg must be a pointer to a struct message than contains
   *          this vertex's id and the termination weight that is
   *          to be sent to the controller.  No modifications are
   *          made to arg before sending it to the controller.
   * POST: The message arg is sent to the controller listening for 
   *          termination weights.  The memory associated with arg
   *          is released back to the heap.
   *****************************************************************/

  friend ostream& operator << (/*in/out*/ ostream& out, 
                               /*in*/ const vertex& s);

 private:

  // PRIVATE MODIFICATION MEMBER FUNCTIONS
  void readNetwork(char* fileName);
  /* Read in the input file that describes the network
   * PRE:  fileName is not null.
   *       fileName is the name of a valid file that describes the 
   *          network
   * POST: The vertex is prepared to start the maxflow 
   *          calculations.
   *       m_edgeList is populated.
   *       m_isSource and m_isSink are set true, if appropriate.
   *       m_excess will be adjusted if this vertex is connected to
   *          the source vertex.
   *****************************************************************/

  // PRIVATE MEMBER VARIABLES for vertex
  map<int,edge> m_edgeMap;
  list<sendFlowData> m_responseList;
  vector<acceptFlowData> m_afdVector;

  bool m_verbose;

  bool m_isSource;
  bool m_isSink;
  int m_sinkId;
  int m_sourceId;

  int m_nextAcceptId;
  int m_currentAcceptId;

  int m_id;
  int m_distance;

  int m_numVertices;

  bool m_terminated, m_mfDirection;

  int m_listenSoc;
  struct sockaddr_in m_controllerAddr;

  int m_maxSocFD;
  fd_set m_fullReadMask;
};
#endif

