// glImage2d.cpp: implementation of the glImage2d class.
//
//////////////////////////////////////////////////////////////////////

#include "glImage2d.h"

#include <stdio.h>
#include <string.h>
#ifdef WIN32
#define strcasecmp(x,y) _strcmpi(x,y)
#endif

#include <png.h>
extern "C" {
#include <jpeglib.h>
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

glImage2d::glImage2d()
{
	m_width=0;
	m_height=0;
}

glImage2d::~glImage2d()
{
}

void glImage2d::Load(const char *const fname)
{
	BYTE *image=0;
    const char *ext=0;
    WORD w,h;
    int i;

    // Find extension
    for(i=strlen(fname),ext=fname+i;i>=0;i--,ext--)
        if(*ext=='.')
        {
            ext++;
            break;
        }

    if(i<0)
    {
#ifdef DEBUG
fprintf(stderr, "glImage2d::Load - "
		        "No extension for file: %s\n",fname);
#endif
        return;
    }

	// ** Determine filetype here **
    if(!strcasecmp(ext,"jpg")||
       !strcasecmp(ext,"jpeg") )
    {
#ifdef DEBUG
fprintf(stderr, "glImage2d::Load - "
		        "Loading JPEG\n");
#endif
	    image=readJPEGImage(fname,w,h);
    }
    else if(!strcasecmp(ext,"png"))
    {
#ifdef DEBUG
fprintf(stderr, "glImage2d::Load - "
		        "Loading PNG\n");
#endif
	    image=readPNGImage(fname,w,h);
    }

    if(!image ||
	   Setup(image,w,h) )
		delete [] image;

	m_width=w;
	m_height=h;

	return;
}

void glImage2d::Resize(const WORD w, const WORD h)
{
	m_width=w;
	m_height=h;
}

BYTE * glImage2d::readJPEGImage(const char *const fname,
								WORD &w,WORD &h         )
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	BYTE channels;
	JSAMPLE *row=0;
	BYTE *image=0,*curimage=0,
		 *currow=0;
	FILE *infile=0;
	unsigned int i,i1;

	// Read JPEG in
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	if ((infile = fopen(fname, "rb")) == NULL)
	{
#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
	            "can't open %s\n", fname);
#endif
	    return 0;
	}

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Initial file read\n");
#endif
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);
	w=cinfo.image_width;
	h=cinfo.image_height;
	channels=cinfo.num_components;

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Width: %i  Height: %i  Channels: %i\n",w,h,channels);
#endif

	// Allocate memory
#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Allocating memory\n");
#endif
	image=new BYTE[w*h*4];
	row=new JSAMPLE[w*channels]; // BYTE also

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Decompressing\n");
#endif
	jpeg_start_decompress(&cinfo);
	curimage=image;
	for(i=0;i<h;i++)
	{
		jpeg_read_scanlines(&cinfo,&row,1);

		// reformat to RGBA
		currow=(BYTE*)row;
		for(i1=0;i1<w;i1++)
			switch(channels)
			{
				case 3: // RGB
					*(curimage++)=*(currow++); // Red
					*(curimage++)=*(currow++); // Green
					*(curimage++)=*(currow++); // Blue
					*(curimage++)=0xFF;        // Alpha
					break;

				case 2: // Treat like greyscale
					currow++;
				case 1: // Greyscale
					*(curimage++)=*currow;     // Red
					*(curimage++)=*currow;     // Green
					*(curimage++)=*(currow++); // Blue
					*(curimage++)=0xFF;        // Alpha
					break;

				default:  // Make white image
					*(curimage++)=0xFF; // Red
					*(curimage++)=0xFF; // Blue
					*(curimage++)=0xFF; // Green
					*(curimage++)=0xFF; // Alpha
			}
	}

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Destroying JPEG objects\n");
#endif
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Closing file handle\n");
#endif
	fclose(infile);

	delete [] row;

#ifdef DEBUG
fprintf(stderr, "glImage2d::readJPEGImage - "
		        "Returning\n");
#endif

	return image;
}

BYTE * glImage2d::readPNGImage(const char *const fname,
							   WORD &w,WORD &h         )
{
	FILE *infile;
	BYTE header[8];
    BYTE *image=0;
    png_structp png_ptr;
    png_infop info_ptr;
    png_infop end_info;
    DWORD width, height;
    int bit_depth, color_type;
    WORD i;

	if ((infile = fopen(fname, "rb")) == NULL)
	{
#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
	            "can't open %s\n", fname);
#endif
	    return 0;
	}

	// *********************************
	// * Copied loosely from Libpng docs
	// *********************************
    fread(header, 1, 8, infile);
    if(png_sig_cmp(header, 0, 8))
        return 0;

    if(!(png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING,
		                                (png_voidp)0,0,0      )))
	{
		fclose(infile);
		return 0;
	}

    if(!(info_ptr=png_create_info_struct(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr,0,0);
		fclose(infile);
        return 0;
    }

    if(!(end_info=png_create_info_struct(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr, &info_ptr,0);
		fclose(infile);
        return 0;
    }

    if (setjmp(png_jmpbuf(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr, &info_ptr,&end_info);
        fclose(infile);
        return 0;
    }

    png_init_io(png_ptr,infile);
    png_set_sig_bytes(png_ptr,8);
    png_set_keep_unknown_chunks(png_ptr,0,0,0);

    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &width, &height,
                 &bit_depth,&color_type,0,0,0       );
    w=width;
    h=height;

	// Format image to RGBA
    if (bit_depth<8)
        png_set_packing(png_ptr);

    if (color_type==PNG_COLOR_TYPE_PALETTE)
        png_set_palette_to_rgb(png_ptr);

    if ((color_type==PNG_COLOR_TYPE_GRAY)||
        (color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
	{
		if(bit_depth<8)
			png_set_gray_1_2_4_to_8(png_ptr);

		png_set_gray_to_rgb(png_ptr);
	}

    if (png_get_valid(png_ptr,info_ptr,PNG_INFO_tRNS))
		png_set_tRNS_to_alpha(png_ptr);

    if (bit_depth == 16)
        png_set_strip_16(png_ptr);

    if ((bit_depth==8)&&(color_type==PNG_COLOR_TYPE_RGB))
		png_set_filler(png_ptr,0xFF,PNG_FILLER_AFTER);
	// Done formatting

    // Read image
#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Width: %i  Height: %i\n",w,h);
#endif

	// Allocate memory
#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Allocating memory\n");
#endif
	image=new BYTE[w*h*4];

#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Decompressing\n");
#endif
	for(i=0;i<h;i++)
        png_read_row(png_ptr, &image[i*w*4], NULL);

    // Clean up
#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Destroying PNG objects\n"  );
#endif
    png_destroy_read_struct(&png_ptr, &info_ptr,&end_info);
#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Closing file handle\n"      );
#endif
    fclose(infile);

#ifdef DEBUG
fprintf(stderr, "glImage2d::readPNGImage - "
		        "Returning\n");
#endif
	return image;
}
