/********************************************************************
** master.h
**-------------------------------------------------------------------
** For the GUIframework, this file contains the master object and
** defines a single instance of it called glApplication
********************************************************************/
#ifndef _MASTER_H_
#define _MASTER_H_

#include <GL/glut.h>
#include <iostream>
#include <map>
#include <list>
#include <stack>
#include <vector>

using namespace std;

/**********************************************************
** Mouse Button State
**---------------------------------------------------------
** Sets the bit fields for a Mouse
** 1st Bit = Left
** 2nd Bit = Right
** 3rd Bit = Middle
**********************************************************/
#define MOUSESTATE_LEFT 1
#define MOUSESTATE_RIGHT 2
#define MOUSESTATE_MIDDLE 4

/**********************************************************
** Required Forward Declaration for the Master Object
**********************************************************/
class guiControl;
class guiContainer;
class guiWindow;
class guiSubWindow;

/**********************************************************
** Simple color structure for holding color information
**********************************************************/
typedef struct _color {
  double r;
  double g;
  double b;
  double a;
}color_t;

/**********************************************************
** This is the Context of the "current control in eye of
** the Master Object
**********************************************************/
struct control_context
{
  int WindowID;
  guiControl *Current;
};

/**********************************************************
** Typedefs
**---------------------------------------------------------
** While only one is needed if futre version this may not 
** be true. In Addition, it makes purpose and the code
** more clear
**
** ControlMap -- Type for a maping of Window ID's
**               to all controls
**  WindowMap -- Type for a maping of Window ID's
**               to only those control which are stand 
**               alone windows
**********************************************************/
typedef map<int, guiControl *, less< int > > ControlMap;
typedef map<int, guiControl *, less< int > > WindowMap;
typedef list<guiControl *> IdleList;
typedef stack<control_context> WindowStack; 

/**********************************************************
** Wrapper Functions
**---------------------------------------------------------
** OpenGL is not OO, so in order to get call back functions
** from within a class to work, they class functions must
** be wrapped in procedural functions
**********************************************************/
void glApplication_Display(void);
void glApplication_Idle(void);
void glApplication_Timer(int value);
void glApplication_Mouse(int key, int state, int x, int y);
void glApplication_Motion(int x, int y);
void glApplication_PassiveMotion(int x, int y);
void glApplication_Keyboard(unsigned char key, int x, int y);
void glApplication_Special(int key, int x, int y);
void glApplication_Visibility(int state);
void glApplication_Reshape(int w, int h);
void glApplication_Entry(int state);

/**********************************************************
** guiMaster
**---------------------------------------------------------
** This is the big cahoona.
**
** Mouse state and current control are Zero upon creation
** Call Init before doing much of anything
** Once you are ready call start.
**---------------------------------------------------------
** NOTE: Dynamically Allocate Controls or suffer weird
**       segfault--I cannot delete objects in static
**       memory
**********************************************************/
class guiMaster
{
  /*
  **These classes are allow to request resources, 
  **so make friends
  */
  friend class guiControl;
  friend class guiContainer;
  friend class guiWindow;
  friend class guiSubWindow;
 
  /*
  ** The wraper function have to get to the real functions
  */
  friend void glApplication_Display(void);
  friend void glApplication_Idle(void);
  friend void glApplication_Timer(int value);
  friend void glApplication_Mouse(int key, int state, int x, int y);
  friend void glApplication_Motion(int x, int y);
  friend void glApplication_PassiveMotion(int x, int y);
  friend void glApplication_Keyboard(unsigned char key, int x, int y);
  friend void glApplication_Special(int key, int x, int y);
  friend void glApplication_Visibility(int state);
  friend void glApplication_Reshape(int w, int h);
  friend void glApplication_Entry(int state);

 public:
  /******************************************************
   ** guiMaster::guiMaster()
   **-----------------------------------------------------
   ** Pre: none
   ** Post: Mouse_Info Returns 0, 
   **       CurrentControl returns NULL
   ******************************************************/
  guiMaster();
  /******************************************************
   ** guiMaster::~guiMaster()
   ** ----------------------------------------------------
   ** Pre: none
   ** Post: recursivly destory each standalone window
   ******************************************************/
  ~guiMaster();

  /******************************************************
   ** void guiMaster::Init(int *argc, char **argc,
   **                          int rate)
   **-----------------------------------------------------
   ** Parameters: argc -- address to argc from main
   **             argv -- argv from main
   ** Pre: none
   ** Post: glutInit(argc,argv) is called
   **       Display Mode is 
   **                    GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA
   **       the idle rate is set to rate.
   ******************************************************/
  void Init(int *argc, char **argv, int rate);
  /******************************************************
   ** void Start(void)
   **-----------------------------------------------------
   ** Start the Glut Main Loop right now
   ******************************************************/
  void Start(void);

  /******************************************************
   ** void guiMaster::SetupWindow(int winID)
   **-----------------------------------------------------
   ** pre: winID is a valid glut Window ID.
   ** post: the frameworks wrapper functions are set as 
   **       the callbacks for the winID
   ******************************************************/
  void SetupWindow(int winID);

  /******************************************************
   ** int guiMaster::Mouse_Info()
   **-----------------------------------------------------
   ** post: returns the the value of the mouse button 
   **       state
   ******************************************************/
  int Mouse_Info();
  /******************************************************
   ** void guiMaster::Set_Application_Idle(void (*func)())
   **-----------------------------------------------------
   ** post: add a general purpose call back to the glut
   **       idle loop.
   ******************************************************/
  void Set_Application_Idle(void (*func)());
  /******************************************************
   ** void guiMaster::SetIdle(int rate)
   **-----------------------------------------------------
   ** pre: Init should be called
   ** post: if int > 0 the IDLE rate is set to this
   **       else the idle loop is disabled.
   ******************************************************/
  void SetIdle(int rate);

  /******************************************************
   ** guiControl *guiMaster::Find_by_ID(int ID)
   **-----------------------------------------------------
   ** pre: ID is a valid glut Window ID
   ** post: the corresponding control is returned.
   ******************************************************/
  guiControl *Find_by_ID(int ID);
  /******************************************************
   ** guiControl *guiMaster:CurrentControl()
   **-----------------------------------------------------
   ** post: if there is a current control return the 
   **       control's pointer else return NULL
   ******************************************************/
  guiControl *CurrentControl();

  bool isStarted();

 protected:
  /*
   ** Controls use this to request a standalone window
   */
    int Request_Window(char name[], guiControl *binding);
    /* 
    ** Controls use this to request a subwindow
    */
    int Request_SubWindow(guiContainer *parent, 
                          guiControl *binding);
    /*
    **  Controls use this to remove display windows
    */
    void Destroy_Window(int ID);
    /*
    ** Controls use this to register an idle function
    */
    void Request_Idletime(guiControl *target);
    /* 
    ** Controls use this to display their idle function
    */
    void Relenquish_Idletime(guiControl *target);
   
    /*
    ** Stack Function
    **------------------------------------------------
    ** Controls when control to control called begin
    ** make sure the right window in place is messy,
    ** so these functions are designed to be called at
    ** the begining of each control functin that needs
    ** a window context. and at the end.  This
    ** "virtually" makes the window ID part of the 
    ** function context, and when a control funciton
    ** returns you can be guaranted that the right
    ** window is active.
    */
    /*Use at the begining of control functions*/
    void NewContext(guiControl *thiscontrol, int ID);
    /*Use at the end of control funciton*/
    void DeleteContext();
    
    
    /*
    ** The "Callback functions
    */
    void Display(void);
    void Idle(void);
    void Mouse(int key, int state, int x, int y);
    void Motion(int x, int y);
    void PassiveMotion(int x, int y);
    void Keyboard(unsigned char key, int x, int y);
    void Special(int key, int x, int y);
    void Visibility(int state);
    void Reshape(int w, int h);
    void Entry(int state);

 private:

    int Started;
    int IdleRate;
    void (*app_idle)();
    int MouseState;
    ControlMap controls;
    WindowMap windows;
    IdleList IdleControls;
    WindowStack controlstack;

};

inline void guiMaster::Init(int *argc, char **argv, int rate){
  glutInit(argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
  SetIdle(rate);
};

inline void guiMaster::Start(void){
  Started = 1;
  glutMainLoop();
};

inline bool guiMaster::isStarted(void){
  return Started;
}

inline guiControl *guiMaster::CurrentControl() {
  if (controlstack.empty()) 
    return 0;
  else
    return (controlstack.top()).Current;
};

inline int guiMaster::Mouse_Info() {
  return MouseState;
}

inline void guiMaster::Set_Application_Idle(void (*func)(void)){
  app_idle = func;
}

extern guiMaster glApplication;

#endif
